"""
Support for parsing Atox format files and converting them into
context-free grammars.
"""


from grammar import *
from xml.sax.handler import ContentHandler


ns = atox_namespace = 'http://hetland.org/atox'


def _adjustAttrNS(attr, url):
    result = {}
    for key, val in attr.items():
        if key[0] is None:
            key = url, key[1]
        result[key] = val
    return result


class FormatError(Exception): pass


class Builder(ContentHandler):

    """
    A SAX content handler used to build Atox parsers from format
    files.
    """
    
    def __init__(self, grammar):
        self._stack = []
        self._chunks = []
        self._symtab = {}
        self._regextab = {}
        self._nameNum = 0
        self.grammar = grammar
        self.grammar.terminals.add(Terminal(name='internal_between')) # XXX

    def newName(self):
        name = 'internal_%i' % self._nameNum
        self._nameNum += 1
        return name

    def checkTag(self, url, name):
        if url == ns:
            if name == 'format': return
            if self.caresAbout((url, name)): return
            raise FormatError('unknown element: %s' % name)

    def startElementNS(self, (url, name), qname, attr):
        self.checkTag(url, name)
        self.attemptRegex()
        self._stack.append((attr, []))

    def addToCurrent(self, item):
        self._stack[-1][-1].append(item)

    def endElementNS(self, (url, name), qname):
        self.checkTag(url, name)
        self.attemptRegex()
        attr, children = self._stack.pop()
        if self.caresAbout((url, name)):
            self.handle((url, name), attr, children)

    def regexFactory(self, text):
        if not text in self._regextab:
            token = RegexToken(self.newName(), text)
            self._regextab[text] = token
        return self._regextab[text]

    def attemptRegex(self):
        if len(self._stack) > 1:
            text = ''.join(self._chunks).strip()
            if text:
                token = self.regexFactory(text)
                self.addToCurrent(token)
        self._chunks = []
        
    def endDocument(self):
        self.grammar.finalize()

    def handlerName(self, name):
        ns, name = name
        if ns != atox_namespace:
            if len(self._stack) == 1:
                self.name = name # XXX
                return 'handleTopSeq'
            name = 'seq'
        return 'handle' + name[0].upper() + name[1:]

    def caresAbout(self, name):
        meth = getattr(self, self.handlerName(name), None)
        return callable(meth)

    def getNamedNonterminal(self, name, prefix='user_'):
        name = prefix + name
        if not name in self._symtab:
            self._symtab[name] = Nonterminal(name=name)
        return self._symtab[name]

    def handle(self, name, attr, children):
        mname = self.handlerName(name)
        method = getattr(self, mname)
        attr = _adjustAttrNS(attr, name[0])
        # XXX Refactor:
        if mname == 'handleSeq' and name[0] != atox_namespace:
            return method(attr, children, tag=name[1])
        else:
            return method(attr, children)

    def handleDef(self, attr, children):
        if not (ns, 'name') in attr:
            raise FormatError('def without a name')
        if not len(attr) == 1:
            raise FormatError('def should have a single attribute')
        if not len(children) > 0:
            raise FormatError('def must contain something')
        lhs = self.getNamedNonterminal(attr[(ns, 'name')])
        self.grammar.newRule(lhs, children)

    def handleMatch(self, attr, children):
        if not (ns, 'name') in attr:
            raise FormatError('match without a name')
        lhs = self.getNamedNonterminal(attr[(ns, 'name')])
        self.addToCurrent(lhs)

    def handleSeq(self, attr, children, tag=None):
        lhs = Nonterminal(name=self.newName())
        self.grammar.newRule(lhs, children, tag)
        self.addToCurrent(lhs)

    # XXX Refactor: Common code with handleSeq.
    # XXX (Could also be collapsed with a sequence with tags; same with rep)
    def handleDel(self, attr, children):
        lhs = Nonterminal(name=self.newName())
        self.grammar.newRule(lhs, children, delete=True)
        self.addToCurrent(lhs)

    def handleAlt(self, attr, children):
        lhs = Nonterminal(name=self.newName())
        for child in children:
            self.grammar.newRule(lhs, [child])
        self.addToCurrent(lhs)

    def handleRep(self, attr, children):
        lhs = Nonterminal(name=self.newName())
        self.grammar.newRule(lhs, [])
        self.grammar.newRule(lhs, [lhs] + children)
        self.addToCurrent(lhs)

    def handlePat(self, attr, children):
        if len(children) != 1:
            raise FormatError('pat should contain a single pattern')
        token = children[0]
        self.addToCurrent(token)

    def handleFill(self, attr, children):
        if children:
            raise FormatError('fill should be an empty tag')
        if attr:
            raise FormatError('fill should have no attributes')
        fill = Fill(name=self.newName())
        self.addToCurrent(fill)

    def handleTopSeq(self, attr, children):
        lhs = Nonterminal(name=self.newName())
        self.grammar.newRule(lhs, children, self.name)
        self.grammar.start = lhs
        self.addToCurrent(lhs)
        
    def characters(self, chars):
        self._chunks.append(chars)
