"""
Representations of context-free grammars. Includes support for
calculating first-sets and follow-sets.
"""


from sets import Set as set

def _update(set1, set2):
    if set2.issubset(set1):
        return False
    set1.update(set2)
    return True


class Nonterminal:

    isTerminal = False

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return '%s' % self.name

    def __repr__(self):
        return 'Nonterminal(name="%s")' % self.name

    def expandFill(self, grammar):
        pass


class Fill(Nonterminal):

    def __repr__(self):
        return 'Fill(name="%s")' % self.name

    def __str__(self):
        return '%s<fill>' % self.name

    def expandFill(self, grammar):
        repeated = Nonterminal(name=self.name + '_rep')
        grammar.newRule(self, [])
        grammar.newRule(self, [self, repeated])
        terminators = grammar.followSet(self)
        legal = grammar.terminals - terminators
        for terminal in legal:
            grammar.newRule(repeated, [terminal])


class Terminal:

    isTerminal = True

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "%s" % self.name

    def __repr__(self):
        return 'Terminal(name="%s")' % self.name


class RegexToken(Terminal):

    def __init__(self, name, expr):
        Terminal.__init__(self, name)
        self.expr = expr

    def __str__(self):
        return "%s(%s)" % (self.name, self.expr)

    def __repr__(self):
        return 'RegexToken(name="%s", expr="%s")' % (self.name, self.expr)


class Rule:

    def __init__(self, lhs, rhs, tag=None, delete=False):
        self.lhs = lhs
        self.rhs = tuple(rhs)
        self.tag = tag
        self.delete = delete

    def __str__(self):
        rhs = ' '.join([str(it) for it in self.rhs])
        if self.tag is not None:
            '%s : %s (tag="%s")' % (self.lhs, rhs, self.tag)
        return '%s : %s' % (self.lhs, rhs)


class GrammarError(Exception): pass


class Grammar:

    def __init__(self):
        self._rules = []
        self.terminals = set()
        self.nonterminals = set()
        self._wrongFirstSets = True
        self._wrongFollowSets = True
        self.start = None

    def finalize(self):
        self.expandFill()
        self.check()

    def expandFill(self):
        for item in self.nonterminals.copy():
            item.expandFill(self)

    def check(self):
        defined = set()
        for rule in self:
            defined.add(rule.lhs)
        if self.start is None:
            raise GrammarError('no start symbol')
        if not self.nonterminals == defined:
            raise GrammarError('undefined nonterminals')

    def addRule(self, rule):
        self._rules.append(rule)
        self.nonterminals.add(rule.lhs)
        for item in rule.rhs:
            if item.isTerminal:
                self.terminals.add(item)
            else:
                self.nonterminals.add(item)
        self._wrongFirstSets = True
        self._wrongFollowSets = True

    def newRule(self, lhs, rhs, tag=None, delete=False):
        self.addRule(Rule(lhs, rhs, tag=tag, delete=delete))

    def _first(self, seq):
        if not seq:
            return set([None])
        if seq[0].isTerminal:
            return set([seq[0]])
        result = self._Fi[seq[0]]
        if not None in result:
            return result
        return (result - set([None])) | self._first(seq[1:])

    def firstSet(self, lhs):
        if self._wrongFirstSets:
            self._calcFirstSets()
        return self._Fi[lhs]

    def followSet(self, lhs):
        if self._wrongFollowSets:
            self._calcFollowSets()
        return self._Fo[lhs]

    def _calcFirstSets(self):
        """
        See http://en.wikipedia.org/wiki/LL_parser for more info.
        """

        Fi = self._Fi = {}
        first = self._first

        for nonterminal in self.nonterminals:
            Fi[nonterminal] = set()
        for rule in self:
            Fi[rule.rhs] = set()

        changed = True
        while changed:
            changed = False
            for rule in self:
                f = first(rule.rhs)
                changed = _update(Fi[rule.rhs], f)
            for rule in self:
                changed = _update(Fi[rule.lhs], Fi[rule.rhs]) or changed

        self._wrongFirstSets = False

    def _calcFollowSets(self):
        """
        See http://en.wikipedia.org/wiki/LL_parser for more info.
        """

        if self._wrongFirstSets:
            self._calcFirstSets()

        Fo = self._Fo = {}
        first = self._first

        for nonterminal in self.nonterminals:
            Fo[nonterminal] = set()

        changed = True
        while changed:
            changed = False
            for rule in self:
                for idx, item in enumerate(rule.rhs):
                    if item.isTerminal: continue
                    f = first(rule.rhs[idx+1:])
                    changed = _update(Fo[item], f - set([None]))
                    if None in f:
                        changed = _update(Fo[item], Fo[rule.lhs]) or changed

        self._wrongFollowSets = False
    
    def __iter__(self):
        return iter(self._rules)

    def __str__(self):
        return '\n'.join([str(rule) for rule in self])
