"""
Support for running the back-end lexer and parser. Includes support
for converting context-free grammars into appropriate input for the
external lexer and parser generators.
"""


from distutils.dist import Distribution
from distutils.core import Extension
from distutils.ccompiler import new_compiler
from glob import glob
from tempfile import mkdtemp
from sets import Set as set
import os, shutil

def flexCode(grammar, moduleName):
    rules = []
    for token in grammar.terminals:
        # XXX Hack:
        if token.name == 'internal_between': continue
        pat, name = token.expr, token.name
        rules.append('<token>%s {\n'
                     '  return handle(%s);\n'
                     '}' % (pat, name))
    body = '\n\n'.join(rules)
    return (FLEX_HEAD % moduleName) + body + FLEX_FOOT

def bisonCode(grammar, moduleName):
    # XXX Only deals with regex tokens and internal_between
    tokenNames = ['internal_between'] + [tok.name for tok in grammar.terminals]
    ruleNames = [rule.lhs.name for rule in grammar]
    head = BISON_HEAD % (' '.join(set(tokenNames)), ' '.join(set(ruleNames) | set(tokenNames)))
    rules = {}
    for rule in grammar: # XXX User rule.tag ...
        alts = rules[rule.lhs] = rules.get(rule.lhs, [])
        alts.append((rule.tag, rule.delete, rule.rhs))
    ruleTexts = []
    for lhs, alts in rules.items():
        rhsTexts = []
        for tag, delete, rhs in alts:
            if not rhs:
                text = '{}'
            else:
                # XXX How safe are these printf actions with GLR?
                # XXX The use of $$ stuff vs. yylval directly; the $$ stuff doesn't
                # XXX seem to work correctly with the current tokenization trick...
                textChunks = []
                for it in rhs:
                    if it.isTerminal: chunk = it.name + ' {output(yylval.text)}'
                    else: chunk = it.name
                    textChunks.append(chunk)
                text = ' '.join(textChunks)
            pre = []
            post = []
            if tag is not None:
                pre.append('output("<%s>")' % tag)
                post.append('output("</%s>")' % tag)
            if delete:
                pre.insert(0, 'start_del()')
                post.append('end_del()')
            if pre or post:
                text = '{%s} %s {%s}' % (';'.join(pre), text, ';'.join(post))
            rhsTexts.append(text)
        rhsText = '\n    | '.join(rhsTexts)
        if lhs is grammar.start:
            ruleTexts.insert(0, '%s:\n      %s;' % (lhs.name, rhsText));
        else:
            ruleTexts.append('%s:\n      %s;' % (lhs.name, rhsText));
    body = '\n\n'.join(ruleTexts)
    foot = BISON_FOOT % ((moduleName,)*6)
    return head + body + foot


def compile(fcode, bcode, outBaseName, dir=None):
   # XXX Drop the "tweening" name
   try:
       if dir is not None:
           startdir = dir
       else:
           startdir = os.getcwd()
       tempdir = mkdtemp()
       os.chdir(tempdir)
    
       l_file_name = outBaseName + '.l'
       y_file_name = outBaseName + '.y'

       l_file = open(l_file_name, 'wb') # Use b to ensure correct newlines (XXX)
       l_file.write(fcode)
       l_file.close()

       y_file = open(y_file_name, 'wb') # Use b to ensure correct newlines (XXX)
       y_file.write(bcode)
       y_file.close()

       l_c_file_name = 'lex.yy.c'
       y_c_file_name = outBaseName + '.tab.c'

       os.system('flex -t %s > %s' % (l_file_name, l_c_file_name))
       os.system('bison -o %s -d %s' % (y_c_file_name, y_file_name)) # Will also create .h file

       dist = Distribution({
           'script_args': ['--quiet', 'build_ext', '--inplace'],
           'ext_modules': [
           Extension(outBaseName, [y_c_file_name, l_c_file_name],
                     extra_compile_args=['-w']) # XXX GCC Specific
           ]
           })
       dist.parse_command_line()
       dist.run_commands()

       # ?!?!?!? Doesn't seem to work in cygwin...
       #tempcompiler = new_compiler()
       #shutil.move(tempcompiler.shared_lib_format % ('tweening', tempcompiler.shared_lib_extension), startdir)

       # XXX Hack:
       libfile = glob('*.dll') or glob('*.so')
       assert len(libfile) == 1
    
       shutil.move(libfile[0], startdir)
   finally:
       # XXX Use atexit?
       shutil.rmtree(tempdir, ignore_errors=True)

# XXX The strdup'ed strings aren't freed anywhere...
# XXX Name checking...
# XXX Wrap errors from flex and bison
# XXX Modify the module method (which calls yyparse) so that it
#     doesn't work directly with sys.stdin

FLEX_HEAD = r"""%%option noyywrap
%%x token tween

%%{

#include <string.h>
#include "%s.tab.h"

char *prev = "";

int current_type = internal_between;
int already_at_eof = 0;
int finished = 0;
int del_level = 0;

void start_del(void) {
  ++del_level;
}

void end_del(void) {
  --del_level;
}

void output(char *text) {
  if (!del_level) printf(text);
}

void catchup(void) {
  char temp;
  temp = *yytext;
  *yytext = '\0';
  yylval.text = strdup(prev);
  *yytext = temp;
  prev = yytext;
}

int catchup_at_eof(void) {
  if (!already_at_eof) {
    yylval.text = strdup(prev);
    already_at_eof = 1;
    if (current_type == internal_between) {
      finished = 1;
    }
    return current_type;
  }
  else if (!finished) {
    yylval.text = "";
    finished = 1;
    return internal_between;
  }
  return 0;
}

int handle(int type) {
  int this_type = current_type;
  catchup();
  current_type = type;
  if (type != internal_between) {
    BEGIN(tween);
  }
  else {
    BEGIN(token);
  }
  return this_type;
}

%%}

%%%%

.|\n {
  prev = yytext;
  yyless(0);
  BEGIN(token);
}

<tween>.|\n {
  yyless(0);
  return handle(internal_between);
}

"""

FLEX_FOOT = r"""

<<EOF>> {
  return catchup_at_eof();
}

<token>.|\n ; /* Don't echo anything */

%%
"""

BISON_HEAD = r"""%%{
#include <stdio.h>
#include <Python.h>

int yyerror(char *s);
int yylex(void);

%%}

%%union {
  char *text;
}

%%glr-parser

%%token %s

%%type <text> %s

%%%%
"""
 
# XXX Add stuff for printing tokens...
BISON_FOOT = r"""
%%%%

/* Change to use Python exception semantics */
/* Should this be a void function? */
int yyerror(char *s) {
  fprintf(stderr, "%%s\n", s);
  return 1;
}

static PyObject *%s_run(PyObject *self, PyObject *args) {
  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  yyparse();
  Py_INCREF(Py_None);
  return Py_None;
}

static PyMethodDef %sMethods[] = {
  {"run", %s_run, METH_VARARGS},
  {NULL, NULL} /* sentinel */
};

void init%s(void) {
  (void) Py_InitModule("%s", %sMethods);
}
"""
